"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var TypeBuilder_1 = require("./TypeBuilder");
var Type_1 = require("./Type");
var Support_1 = require("./Support");
var TypeAttributes_1 = require("./TypeAttributes");
function intersectionMembersRecursively(intersection) {
    var types = [];
    var attributes = TypeAttributes_1.emptyTypeAttributes;
    function process(t) {
        if (t instanceof Type_1.IntersectionType) {
            attributes = TypeAttributes_1.combineTypeAttributes([attributes, t.getAttributes()]);
            t.members.forEach(process);
        }
        else if (t.kind !== "any") {
            types.push(t);
        }
        else {
            attributes = TypeAttributes_1.combineTypeAttributes([attributes, t.getAttributes()]);
        }
    }
    process(intersection);
    return [immutable_1.OrderedSet(types), attributes];
}
function canResolve(t) {
    var members = intersectionMembersRecursively(t)[0];
    if (members.size <= 1)
        return true;
    return members.every(function (m) { return !(m instanceof Type_1.UnionType) || m.isCanonical; });
}
var IntersectionAccumulator = /** @class */ (function () {
    function IntersectionAccumulator() {
        // We allow only either maps, classes, or neither.  States:
        //
        // 1. Start: No types seen yet, both are allowed, _mapValueTypes is
        //    the empty set, _classProperties is undefined.
        // 2. At least one type seen, all of them can be maps: _mapValueTypes
        //    is a non-empty set, _classProperties is undefined.
        // 3. All types seen can be maps or classes, at least one of them
        //    can only be a class: Maps are not allowed anymore, but classes
        //    are.  _mapValueTypes is undefined, _classProperties is defined.
        // 4. At least one type seen that can't be map or class: Neither map
        //    nor class is allowed anymore.  _mapValueTypes and _classProperties
        //    are both undefined.
        this._mapValueTypes = immutable_1.OrderedSet();
    }
    IntersectionAccumulator.prototype.updatePrimitiveStringTypes = function (members) {
        var types = members.filter(function (t) { return Type_1.isPrimitiveStringTypeKind(t.kind); });
        var kinds = types.map(function (t) { return t.kind; });
        if (this._primitiveStringTypes === undefined) {
            this._primitiveStringTypes = kinds;
            return;
        }
        if (members.find(function (t) { return t instanceof Type_1.StringType; }) === undefined) {
            this._primitiveStringTypes = this._primitiveStringTypes.intersect(kinds);
        }
    };
    IntersectionAccumulator.prototype.updateOtherPrimitiveTypes = function (members) {
        var types = members.filter(function (t) { return Type_1.isPrimitiveTypeKind(t.kind) && !Type_1.isPrimitiveStringTypeKind(t.kind); });
        var kinds = types.map(function (t) { return t.kind; });
        if (this._otherPrimitiveTypes === undefined) {
            this._otherPrimitiveTypes = kinds;
            return;
        }
        var haveNumber = this._otherPrimitiveTypes.find(Type_1.isNumberTypeKind) !== undefined &&
            kinds.find(Type_1.isNumberTypeKind) !== undefined;
        this._otherPrimitiveTypes = this._otherPrimitiveTypes.intersect(kinds);
        if (haveNumber && this._otherPrimitiveTypes.find(Type_1.isNumberTypeKind) === undefined) {
            // One set has integer, the other has double.  The intersection
            // of that is integer.
            this._otherPrimitiveTypes = this._otherPrimitiveTypes.add("integer");
        }
    };
    IntersectionAccumulator.prototype.updateEnumCases = function (members) {
        if (members.find(function (t) { return t instanceof Type_1.StringType; }) !== undefined) {
            return;
        }
        var newCases = (_a = immutable_1.OrderedSet()).union.apply(_a, members.map(function (t) { return (t instanceof Type_1.EnumType ? t.cases : immutable_1.OrderedSet()); }).toArray());
        if (this._enumCases === undefined) {
            this._enumCases = newCases;
        }
        else {
            this._enumCases = this._enumCases.intersect(newCases);
        }
        var _a;
    };
    IntersectionAccumulator.prototype.updateArrayItemTypes = function (members) {
        if (this._arrayItemTypes === false)
            return;
        var maybeArray = members.find(function (t) { return t instanceof Type_1.ArrayType; });
        if (maybeArray === undefined) {
            this._arrayItemTypes = false;
            return;
        }
        if (this._arrayItemTypes === undefined) {
            this._arrayItemTypes = immutable_1.OrderedSet();
        }
        this._arrayItemTypes = this._arrayItemTypes.add(maybeArray.items);
    };
    IntersectionAccumulator.prototype.updateMapValueTypesAndClassProperties = function (members) {
        function makeProperties() {
            if (maybeClass === undefined)
                return Support_1.panic("Didn't we just check for this?");
            return maybeClass.properties.map(function (cp) { return new Type_1.GenericClassProperty(immutable_1.OrderedSet([cp.type]), cp.isOptional); });
        }
        var maybeClass = members.find(function (t) { return t instanceof Type_1.ClassType; });
        var maybeMap = members.find(function (t) { return t instanceof Type_1.MapType; });
        Support_1.assert(maybeClass === undefined || maybeMap === undefined, "Can't have both class and map type in a canonical union");
        if (maybeMap === undefined && maybeClass === undefined) {
            // Moving to state 4.
            this._mapValueTypes = undefined;
            this._classProperties = undefined;
            return;
        }
        if (this._mapValueTypes !== undefined) {
            // We're in state 1 or 2.
            Support_1.assert(this._classProperties === undefined, "One of _mapValueTypes and _classProperties must be undefined");
            if (maybeMap !== undefined) {
                // Moving to state 2.
                this._mapValueTypes = this._mapValueTypes.add(maybeMap.values);
            }
            else {
                // Moving to state 3.
                this._mapValueTypes = undefined;
                this._classProperties = makeProperties();
            }
        }
        else if (this._classProperties !== undefined) {
            // We're in state 3.
            if (maybeMap !== undefined) {
                this._classProperties = this._classProperties.map(function (cp) { return new Type_1.GenericClassProperty(cp.typeData.add(maybeMap.values), cp.isOptional); });
            }
            else {
                // Staying in state 3.
                if (maybeClass === undefined)
                    return Support_1.panic("Didn't we just check for this?");
                this._classProperties = this._classProperties.mergeWith(function (cp1, cp2) {
                    return new Type_1.GenericClassProperty(cp1.typeData.union(cp2.typeData), cp1.isOptional || cp2.isOptional);
                }, makeProperties());
            }
        }
        else {
            // We're in state 4.  No way out of state 4.
        }
        Support_1.assert(this._mapValueTypes === undefined || this._classProperties === undefined, "We screwed up our sacred state machine.");
    };
    IntersectionAccumulator.prototype.addAny = function (_t) {
        // "any" doesn't change the types at all
        // FIXME: just add attributes
    };
    IntersectionAccumulator.prototype.addUnionSet = function (members) {
        this.updatePrimitiveStringTypes(members);
        this.updateOtherPrimitiveTypes(members);
        this.updateEnumCases(members);
        this.updateArrayItemTypes(members);
        this.updateMapValueTypesAndClassProperties(members);
    };
    IntersectionAccumulator.prototype.addUnion = function (u) {
        this.addUnionSet(u.members);
    };
    IntersectionAccumulator.prototype.addType = function (t) {
        var _this = this;
        // FIXME: We're very lazy here.  We're supposed to keep type
        // attributes separately for each type kind, but we collect
        // them all together and return them as attributes for the
        // overall result type.
        var attributes = t.getAttributes();
        Type_1.matchTypeExhaustive(t, function (_noneType) {
            return Support_1.panic("There shouldn't be a none type");
        }, function (anyType) { return _this.addAny(anyType); }, function (nullType) { return _this.addUnionSet(immutable_1.OrderedSet([nullType])); }, function (boolType) { return _this.addUnionSet(immutable_1.OrderedSet([boolType])); }, function (integerType) { return _this.addUnionSet(immutable_1.OrderedSet([integerType])); }, function (doubleType) { return _this.addUnionSet(immutable_1.OrderedSet([doubleType])); }, function (stringType) { return _this.addUnionSet(immutable_1.OrderedSet([stringType])); }, function (arrayType) { return _this.addUnionSet(immutable_1.OrderedSet([arrayType])); }, function (classType) { return _this.addUnionSet(immutable_1.OrderedSet([classType])); }, function (mapType) { return _this.addUnionSet(immutable_1.OrderedSet([mapType])); }, function (enumType) { return _this.addUnionSet(immutable_1.OrderedSet([enumType])); }, function (unionType) { return _this.addUnion(unionType); }, function (dateType) { return _this.addUnionSet(immutable_1.OrderedSet([dateType])); }, function (timeType) { return _this.addUnionSet(immutable_1.OrderedSet([timeType])); }, function (dateTimeType) { return _this.addUnionSet(immutable_1.OrderedSet([dateTimeType])); });
        return attributes;
    };
    Object.defineProperty(IntersectionAccumulator.prototype, "arrayData", {
        get: function () {
            if (this._arrayItemTypes === undefined || this._arrayItemTypes === false) {
                return Support_1.panic("This should not be called if the type can't be an array");
            }
            return this._arrayItemTypes;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(IntersectionAccumulator.prototype, "mapData", {
        get: function () {
            return this._mapValueTypes;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(IntersectionAccumulator.prototype, "classData", {
        get: function () {
            return this._classProperties;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(IntersectionAccumulator.prototype, "enumCases", {
        get: function () {
            return Support_1.defined(this._enumCases).toArray();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(IntersectionAccumulator.prototype, "enumCaseMap", {
        get: function () {
            var caseMap = {};
            Support_1.defined(this._enumCases).forEach(function (n) { return (caseMap[n] = 1); });
            return caseMap;
        },
        enumerable: true,
        configurable: true
    });
    IntersectionAccumulator.prototype.getMemberKinds = function () {
        var kinds = Support_1.defined(this._primitiveStringTypes).union(Support_1.defined(this._otherPrimitiveTypes));
        if (this._enumCases !== undefined && this._enumCases.size > 0) {
            kinds = kinds.add("enum");
        }
        if (immutable_1.OrderedSet.isOrderedSet(this._arrayItemTypes)) {
            kinds = kinds.add("array");
        }
        if (this._mapValueTypes !== undefined) {
            kinds = kinds.add("map");
        }
        else if (this._classProperties !== undefined) {
            kinds = kinds.add("class");
        }
        return kinds.toOrderedMap().map(function (_) { return TypeAttributes_1.emptyTypeAttributes; });
    };
    return IntersectionAccumulator;
}());
var IntersectionUnionBuilder = /** @class */ (function (_super) {
    __extends(IntersectionUnionBuilder, _super);
    function IntersectionUnionBuilder() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this._createdNewIntersections = false;
        return _this;
    }
    IntersectionUnionBuilder.prototype.makeIntersection = function (members, attributes) {
        var _this = this;
        var reconstitutedMembers = members.map(function (t) { return _this.typeBuilder.reconstituteTypeRef(t.typeRef); });
        var first = Support_1.defined(reconstitutedMembers.first());
        if (reconstitutedMembers.size === 1) {
            this.typeBuilder.addAttributes(first, attributes);
            return first;
        }
        this._createdNewIntersections = true;
        return this.typeBuilder.getUniqueIntersectionType(attributes, reconstitutedMembers);
    };
    Object.defineProperty(IntersectionUnionBuilder.prototype, "createdNewIntersections", {
        get: function () {
            return this._createdNewIntersections;
        },
        enumerable: true,
        configurable: true
    });
    IntersectionUnionBuilder.prototype.makeEnum = function (cases, _counts, typeAttributes, forwardingRef) {
        return this.typeBuilder.getEnumType(typeAttributes, immutable_1.OrderedSet(cases), forwardingRef);
    };
    IntersectionUnionBuilder.prototype.makeClass = function (maybeProperties, maybeMapValueTypes, typeAttributes, forwardingRef) {
        var _this = this;
        if (maybeProperties !== undefined) {
            Support_1.assert(maybeMapValueTypes === undefined);
            var tref = this.typeBuilder.getUniqueClassType(typeAttributes, true, undefined, forwardingRef);
            // FIXME: attributes
            var properties = maybeProperties.map(function (cp) { return new Type_1.ClassProperty(_this.makeIntersection(cp.typeData, immutable_1.Map()), cp.isOptional); });
            this.typeBuilder.setClassProperties(tref, properties);
            return tref;
        }
        else if (maybeMapValueTypes !== undefined) {
            // FIXME: attributes
            var valuesType = this.makeIntersection(maybeMapValueTypes, immutable_1.Map());
            var mapType = this.typeBuilder.getMapType(valuesType, forwardingRef);
            this.typeBuilder.addAttributes(mapType, typeAttributes);
            return mapType;
        }
        else {
            return Support_1.panic("Either classes or maps must be given");
        }
    };
    IntersectionUnionBuilder.prototype.makeArray = function (arrays, typeAttributes, forwardingRef) {
        // FIXME: attributes
        var itemsType = this.makeIntersection(arrays, immutable_1.Map());
        var tref = this.typeBuilder.getArrayType(itemsType, forwardingRef);
        this.typeBuilder.addAttributes(tref, typeAttributes);
        return tref;
    };
    return IntersectionUnionBuilder;
}(TypeBuilder_1.UnionBuilder));
function resolveIntersections(graph, stringTypeMapping) {
    var needsRepeat = false;
    function replace(types, builder, forwardingRef) {
        Support_1.assert(types.size === 1);
        var _a = intersectionMembersRecursively(Support_1.defined(types.first())), members = _a[0], intersectionAttributes = _a[1];
        if (members.isEmpty()) {
            return builder.getPrimitiveType("any", forwardingRef);
        }
        if (members.size === 1) {
            var single = builder.reconstituteType(Support_1.defined(members.first()), forwardingRef);
            builder.addAttributes(single, intersectionAttributes);
            return single;
        }
        var accumulator = new IntersectionAccumulator();
        var extraAttributes = TypeAttributes_1.makeTypeAttributesInferred(TypeAttributes_1.combineTypeAttributes(members.map(function (t) { return accumulator.addType(t); }).toArray()));
        var attributes = TypeAttributes_1.combineTypeAttributes([intersectionAttributes, extraAttributes]);
        var unionBuilder = new IntersectionUnionBuilder(builder);
        var tref = unionBuilder.buildUnion(accumulator, true, attributes, forwardingRef);
        if (unionBuilder.createdNewIntersections) {
            needsRepeat = true;
        }
        return tref;
    }
    // FIXME: We need to handle intersections that resolve to the same set of types.
    // See for example the intersections-nested.schema example.
    var intersections = graph.allTypesUnordered().filter(function (t) { return t instanceof Type_1.IntersectionType; });
    if (intersections.isEmpty()) {
        return [graph, true];
    }
    var resolvableIntersections = intersections.filter(canResolve);
    if (resolvableIntersections.isEmpty()) {
        return [graph, false];
    }
    var groups = resolvableIntersections.map(function (i) { return [i]; }).toArray();
    graph = graph.rewrite(stringTypeMapping, false, groups, replace);
    // console.log(`resolved ${resolvableIntersections.size} of ${intersections.size} intersections`);
    return [graph, !needsRepeat && intersections.size === resolvableIntersections.size];
}
exports.resolveIntersections = resolveIntersections;
